const NUMBER_THUMBNAILS = 9;
const THUMB_FOLDER = "SDThumbs";

// Global vars
var gPref = Components.classes["@mozilla.org/preferences-service;1"]
               .getService(Components.interfaces.nsIPrefService)
               .QueryInterface(Components.interfaces.nsIPrefBranch2);
var speedDialBundle;

// Configuration parameters
var thumbnailTooltipTitle = true;
var thumbnailTooltipURL = true;

// Globarl vars
var thumbnailCanvasWidth = 0;
var thumbnailCanvasHeight = 0;
var updateTimer = null;
var updateItems = new Array(9);
var boardWidth;
var boardHeight;
var selectionStart;
var selectionEnd;
var thumbnailDragDropExclude;

function speeddialStartup() {
  // Load bundle
  speedDialBundle = document.getElementById("bundle_speeddial");

  
  var innerBox = document.getElementById("speeddialBox");
  boardWidth = innerBox.boxObject.width;
  boardHeight = innerBox.boxObject.height;

  restoreCanvasSize();
  
  // Register observers
  SpeedDialPrefObserver.addPrefObserver();
  window.addEventListener("resize", sizeChanged, false);
  window.addEventListener("mousedown", windowMouseDown, true);

  // Load settings
  thumbnailTooltipTitle = gPref.getBoolPref("extensions.speeddial.thumbnailTooltipTitle");
  thumbnailTooltipURL = gPref.getBoolPref("extensions.speeddial.thumbnailTooltipURL");
  thumbnailDragDropExclude = gPref.getIntPref("extensions.speeddial.thumbnailDragDropExclude");

  
  if (gPref.getBoolPref("extensions.speeddial.focusContentOnLoad")) {
    document.getElementById("speeddialBox").focus();
  }

  if (gPref.getBoolPref("extensions.speeddial.clearURLBarOnLoad")) {
    // Save selection ranges
    var parentWindow = getParentWindow();
    var urlBar = parentWindow.document.getElementById("urlbar");
    selectionStart = urlBar.selectionStart;
    selectionEnd = urlBar.selectionEnd;
  }

  setTimeout(speeddialAfterLoad, 0);
}

function speeddialAfterLoad() {
  var parentWindow = getParentWindow();
  var urlBar = parentWindow.document.getElementById("urlbar");

  if (gPref.getBoolPref("extensions.speeddial.clearURLBarOnLoad")) {
    urlBar.editor.undo(2);
    // Restore selection ranges
    urlBar.setSelectionRange(selectionStart, selectionEnd);
  }

  if (gPref.getBoolPref("extensions.speeddial.focusURLBarOnLoad")) {
    setTimeout(speeddialFocusURLBar, 0);
  }
}

function checkThumbnailDragDropExcludeCanvas(event) {
  if ((thumbnailDragDropExclude & 1) > 0) {
    event.stopPropagation();
  }
}

function speeddialFocusURLBar() {
  var parentWindow = getParentWindow();
  var urlBar = parentWindow.document.getElementById("urlbar");
  urlBar.focus();
  urlBar.select();
}

function speeddialUnload() {
  // Unregister observers
  SpeedDialPrefObserver.removePrefObserver();
  window.removeEventListener("resize", sizeChanged, false);
  window.removeEventListener("mousedown", windowMouseDown, true);
}

function restoreCanvasSize() {
  var canvasBox = document.getBoxObjectFor(document.getElementById("speeddial1").getCanvas());
  thumbnailCanvasWidth = canvasBox.width;
  thumbnailCanvasHeight = canvasBox.height;

  // Update thumbnails
  for (var c=1; c<=NUMBER_THUMBNAILS; c++) {
    var currentThumbnail = document.getElementById("speeddial" + c);
    currentThumbnail.updateThumbnail();
    updateItems[c] = false;
  }
}

function nullifyCanvasSize() {
  for (var c=1; c<=NUMBER_THUMBNAILS; c++) {
    var currentThumbnail = document.getElementById("speeddial" + c);
    currentThumbnail.nullifyCanvas();
  }
}

function updatePendingItems() {
  for (var c=1; c<=NUMBER_THUMBNAILS; c++) {
    if (updateItems[c]) {
      var currentThumbnail = document.getElementById("speeddial" + c);
      currentThumbnail.updateThumbnail();
      updateItems[c] = false;
    }
  }
}

function openURL(url, where) {
  if (!url) return;

  var parentWindow = getParentWindow();
  if (where == "current") {
    document.location = url;
  } else {
    parentWindow.openUILinkIn(url, where);
  }
}

function editThumbnail(thumbNumber) {
  openDialog("chrome://speeddial/content/editDial.xul", "",
             "centerscreen,chrome,dialog,resizable,dependent", [ thumbNumber ]);
}

function refreshThumbnail(thumbNumber) {
  gPref.setBoolPref("extensions.speeddial.thumbnail-" + thumbNumber + "-manualrefresh", true);
}

function deassignThumbnail(thumbNumber) {
  removeThumbnailImage(thumbNumber);

  if (gPref.prefHasUserValue("extensions.speeddial.thumbnail-" + thumbNumber + "-url")) {
    gPref.clearUserPref("extensions.speeddial.thumbnail-" + thumbNumber + "-url");
  }

  if (gPref.prefHasUserValue("extensions.speeddial.thumbnail-" + thumbNumber + "-icon")) {
    gPref.clearUserPref("extensions.speeddial.thumbnail-" + thumbNumber + "-icon");
  }

  if (gPref.prefHasUserValue("extensions.speeddial.thumbnail-" + thumbNumber + "-label")) {
    gPref.clearUserPref("extensions.speeddial.thumbnail-" + thumbNumber + "-label");
  }

  if (gPref.prefHasUserValue("extensions.speeddial.thumbnail-" + thumbNumber + "-lastsaved")) {
    gPref.clearUserPref("extensions.speeddial.thumbnail-" + thumbNumber + "-lastsaved");
  }

  if (gPref.prefHasUserValue("extensions.speeddial.thumbnail-" + thumbNumber + "-dynamictitle")) {
    gPref.clearUserPref("extensions.speeddial.thumbnail-" + thumbNumber + "-dynamictitle");
  }

  if (gPref.prefHasUserValue("extensions.speeddial.thumbnail-" + thumbNumber + "-refreshinterval")) {
    gPref.clearUserPref("extensions.speeddial.thumbnail-" + thumbNumber + "-refreshinterval");
  }

  if (gPref.prefHasUserValue("extensions.speeddial.thumbnail-" + thumbNumber + "-manualrefresh")) {
    gPref.clearUserPref("extensions.speeddial.thumbnail-" + thumbNumber + "-manualrefresh");
  }
}

function showingThumbnailTooltip(event) {
  event.stopPropagation();

  var currentTitle = document.getElementById("thumbnailTooltipTitle");
  var currentURL = document.getElementById("thumbnailTooltipURL");
  var showTooltip = false;

  var targetNode = document.tooltipNode;

  if (thumbnailTooltipTitle && targetNode.hasAttribute("thumbnailTitle")) {
    currentTitle.removeAttribute("hidden");
    currentTitle.value = targetNode.getAttribute("thumbnailTitle");
    showTooltip = true;
  } else {
    currentTitle.setAttribute("hidden", true);
  }

  if (thumbnailTooltipURL && targetNode.hasAttribute("thumbnailURL")) {
    currentURL.removeAttribute("hidden");
    currentURL.value = targetNode.getAttribute("thumbnailURL");
    showTooltip = true;
  } else {
    currentURL.setAttribute("hidden", true);
  }
  return showTooltip;
}

function getParentWindow() {
  if (window.opener) {
    return window.opener;
  } else {
    return QueryInterface(Components.interfaces.nsIInterfaceRequestor).getInterface(Components.interfaces.nsIWebNavigation).QueryInterface(Components.interfaces.nsIDocShellTreeItem).rootTreeItem.QueryInterface(Components.interfaces.nsIInterfaceRequestor).getInterface(Components.interfaces.nsIDOMWindow);
  }
}

function getThumbnailImageURL(thumbnailNumber) {
  var file = Components.classes["@mozilla.org/file/directory_service;1"]
                     .getService(Components.interfaces.nsIProperties)
                     .get("ProfD", Components.interfaces.nsIFile);
  file.append(THUMB_FOLDER);
  if (!gPref.prefHasUserValue("extensions.speeddial.thumbnail-" + thumbnailNumber + "-lastsaved")) {
    file.append("thumbnail-" + thumbnailNumber + ".png");
  } else {
    file.append("thumbnail-" + thumbnailNumber + "-" + gPref.getCharPref("extensions.speeddial.thumbnail-" + thumbnailNumber + "-lastsaved") + ".png");
  }
  if (file.exists() && !file.isDirectory()) {
    var ios = Components.classes["@mozilla.org/network/io-service;1"]
                    .getService(Components.interfaces.nsIIOService);
    return ios.newFileURI(file).spec;
  } else {
    return null;
  }
}

function removeThumbnailImage(thumbnailNumber) {
  var file = Components.classes["@mozilla.org/file/directory_service;1"]
                     .getService(Components.interfaces.nsIProperties)
                     .get("ProfD", Components.interfaces.nsIFile);
  file.append(THUMB_FOLDER);
  if (gPref.prefHasUserValue("extensions.speeddial.thumbnail-" + thumbnailNumber + "-lastsaved")) {
    file.append("thumbnail-" + thumbnailNumber + "-" + gPref.getCharPref("extensions.speeddial.thumbnail-" + thumbnailNumber + "-lastsaved") + ".png");
  } else {
    file.append("thumbnail-" + thumbnailNumber + ".png");
  }
  if (file.exists() && !file.isDirectory()) {
    file.remove(false);
  }
}

function moveThumbnailImage(originalNumber, destinationNumber) {
  var file = Components.classes["@mozilla.org/file/directory_service;1"]
                     .getService(Components.interfaces.nsIProperties)
                     .get("ProfD", Components.interfaces.nsIFile);
  file.append(THUMB_FOLDER);
  var thumbFolder = file.clone();

  if (gPref.prefHasUserValue("extensions.speeddial.thumbnail-" + originalNumber + "-lastsaved")) {
    file.append("thumbnail-" + originalNumber + "-" + gPref.getCharPref("extensions.speeddial.thumbnail-" + originalNumber + "-lastsaved") + ".png");
  } else {
    file.append("thumbnail-" + originalNumber + ".png");
  }

  var destFileName;
  if (gPref.prefHasUserValue("extensions.speeddial.thumbnail-" + destinationNumber + "-lastsaved")) {
    destFileName = "thumbnail-" + destinationNumber + "-" + gPref.getCharPref("extensions.speeddial.thumbnail-" + destinationNumber + "-lastsaved") + ".png";
  } else {
    destFileName = "thumbnail-" + destinationNumber + ".png";
  }

  if (file.exists() && !file.isDirectory()) {
    file.moveTo(thumbFolder, destFileName);
  }
}

function copyThumbnailImage(originalNumber, destinationNumber) {
  var file = Components.classes["@mozilla.org/file/directory_service;1"]
                     .getService(Components.interfaces.nsIProperties)
                     .get("ProfD", Components.interfaces.nsIFile);
  file.append(THUMB_FOLDER);
  var thumbFolder = file.clone();

  if (gPref.prefHasUserValue("extensions.speeddial.thumbnail-" + originalNumber + "-lastsaved")) {
    file.append("thumbnail-" + originalNumber + "-" + gPref.getCharPref("extensions.speeddial.thumbnail-" + originalNumber + "-lastsaved") + ".png");
  } else {
    file.append("thumbnail-" + originalNumber + ".png");
  }

  var destFileName;
  if (gPref.prefHasUserValue("extensions.speeddial.thumbnail-" + destinationNumber + "-lastsaved")) {
    destFileName = "thumbnail-" + destinationNumber + "-" + gPref.getCharPref("extensions.speeddial.thumbnail-" + destinationNumber + "-lastsaved") + ".png";
  } else {
    destFileName = "thumbnail-" + destinationNumber + ".png";
  }

  if (file.exists() && !file.isDirectory()) {
    file.copyTo(thumbFolder, destFileName);
  }
}


function getThumbnailCanvasWidth() {
  return thumbnailCanvasWidth;
}

function getThumbnailCanvasHeight() {
  return thumbnailCanvasHeight;
}

function checkShouldShowContextMenu(thumbNumber) {
  return gPref.prefHasUserValue("extensions.speeddial.thumbnail-" + thumbNumber + "-url");
}

function sizeChanged() {
/*
  var innerBox = document.getElementById("speeddialBox");
  
  if ((innerBox.boxObject.width < boardWidth) || (innerBox.boxObject.height < boardHeight)) {
    boardWidth = innerBox.boxObject.width;
    boardHeight = innerBox.boxObject.height;
    nullifyCanvasSize();
    setTimeout(restoreCanvasSize, 0);
  } else {
    restoreCanvasSize();
  }
*/
  nullifyCanvasSize();
  setTimeout(restoreCanvasSize, 0);
}

function windowMouseDown(el) {
  if (el.button == 1) {
    el.stopPropagation();
//    el.preventDefault();
  }
}

var SpeedDialPrefObserver = 
{
  prefObserver : {
    observe: function(subject, topic, data) {
    // subject is the nsIPrefBranch we're observing (after appropriate QI)
    // data is the name of the pref that's been changed (relative to subject)
      if (topic == "nsPref:changed") {
        if (data.substr(0, "extensions.speeddial.thumbnail-".length) == "extensions.speeddial.thumbnail-") {
          // A thumbnail is updated!
          var number = data.substr("extensions.speeddial.thumbnail-".length, 1);
          updateItems[number] = true;

          if (updateTimer == null) {
            updateTimer = Components.classes["@mozilla.org/timer;1"].createInstance(Components.interfaces.nsITimer);
          }
          updateTimer.cancel();
          updateTimer.initWithCallback(SpeedDialPrefObserver.prefObserver, 100, Components.interfaces.nsITimer.TYPE_ONE_SHOT);
        }
      } else {
        if (data == "extensions.speeddial.thumbnailTooltipTitle") {
          thumbnailTooltipTitle = gPref.getBoolPref("extensions.speeddial.thumbnailTooltipTitle");
        } else if (data == "extensions.speeddial.thumbnailTooltipTitle") {
          thumbnailTooltipURL = gPref.getBoolPref("extensions.speeddial.thumbnailTooltipURL");
        } else if (data == "extensions.speeddial.thumbnailDragDropExclude") {
          thumbnailDragDropExclude = gPref.getIntPref("extensions.speeddial.thumbnailDragDropExclude");
        }
      }
    },

    QueryInterface : function (aIID) {
      if (aIID.equals(Components.interfaces.nsIObserver) || 
      aIID.equals(Components.interfaces.nsITimerCallback) ||
      aIID.equals(Components.interfaces.nsISupports) ||
      aIID.equals(Components.interfaces.nsISupportsWeakReference))
        return this;
      throw Components.results.NS_NOINTERFACE;
    },

    notify: function(timer) {
      updatePendingItems();
    }
  },

  addPrefObserver : function () {
    if (!gPref) return;
    gPref.addObserver("extensions.speeddial.", SpeedDialPrefObserver.prefObserver, true);
  },

  removePrefObserver : function () {
    if (!gPref) return;
    gPref.removeObserver("extensions.speeddial.", SpeedDialPrefObserver.prefObserver);
  }
};

function ThumbnailImageLoader(thumbnail, canvas, image) {
  this.thumbnail = thumbnail;
  this.canvas = canvas;
  this.image = image;
  image.addEventListener("load", this, false);
  this.thumbnail.loader = this;
}

ThumbnailImageLoader.prototype = {
  canvas: null,
  image: null,
  thumbnail: null,

  handleEvent: function(event) {
    this.canvas.cachedImage = this.image;

    var ctx = this.canvas.getContext('2d');
    ctx.fillStyle = "white";
    ctx.fillRect(0, 0, this.canvas.width, this.canvas.height);
    ctx.drawImage(this.image, 0, 0, this.canvas.width, this.image.height * this.canvas.width / this.image.width);
/*
    var sHeight = this.canvas.width * this.image.height / this.image.width;
    ctx.drawImage(this.image, 0, 0, this.image.width, sHeight, 0, 0, this.canvas.width, this.canvas.height);
*/
    this.image.removeEventListener("load", this, false);
    this.thumbnail.loader = undefined;
  }
}

// Drag & drop
var thumbnailObserver = {
  onDragStart: function (evt,transferData,action) {
    transferData.data = createTransferDataForEntry(evt.currentTarget.getAttribute("speeddial"));
    if (transferData.data == null)
      return false;
  },

  getSupportedFlavours : function () {
    var flavours = new FlavourSet();
    flavours.appendFlavour("text/x-speeddial-entry");
    flavours.appendFlavour("text/x-moz-url");
    flavours.appendFlavour("application/x-moz-file", "nsIFile");
    return flavours;
  },

  onDragOver: function (evt,flavour,session) {
  },

  onDragExit: function (aEvent, aDragSession) {
  },

  onDrop: function (aEvent,dropdata,session) {
try {
    const kDSIID      = Components.interfaces.nsIDragService;
    const kCopyAction = kDSIID.DRAGDROP_ACTION_COPY + kDSIID.DRAGDROP_ACTION_LINK;
    var moveIfPossible = !(session.dragAction & kCopyAction);

    var targetData = getSelectionFromXferData(session);
    var targetDial = aEvent.currentTarget.getAttribute("speeddial");

    if (targetData.type == "speedDialId") {
      var originalDial = targetData.value;
      // Remove old entry
      deassignThumbnail(targetDial);

      // Copy data
      if (gPref.prefHasUserValue("extensions.speeddial.thumbnail-" + originalDial + "-url")) {
        gPref.setCharPref("extensions.speeddial.thumbnail-" + targetDial + "-url", gPref.getCharPref("extensions.speeddial.thumbnail-" + originalDial + "-url"));
      }

      if (gPref.prefHasUserValue("extensions.speeddial.thumbnail-" + originalDial + "-icon")) {
        gPref.setCharPref("extensions.speeddial.thumbnail-" + targetDial + "-icon", gPref.getCharPref("extensions.speeddial.thumbnail-" + originalDial + "-icon"));
      }

      if (gPref.prefHasUserValue("extensions.speeddial.thumbnail-" + originalDial + "-label")) {
        gPref.setComplexValue("extensions.speeddial.thumbnail-" + targetDial + "-label", Components.interfaces.nsISupportsString, gPref.getComplexValue("extensions.speeddial.thumbnail-" + originalDial + "-label", Components.interfaces.nsISupportsString));
      }

      if (gPref.prefHasUserValue("extensions.speeddial.thumbnail-" + originalDial + "-lastsaved")) {
        gPref.setCharPref("extensions.speeddial.thumbnail-" + targetDial + "-lastsaved", gPref.getCharPref("extensions.speeddial.thumbnail-" + originalDial + "-lastsaved"));
      }

      if (gPref.prefHasUserValue("extensions.speeddial.thumbnail-" + originalDial + "-dynamictitle")) {
        gPref.setBoolPref("extensions.speeddial.thumbnail-" + targetDial + "-dynamictitle", gPref.getBoolPref("extensions.speeddial.thumbnail-" + originalDial + "-dynamictitle"));
      }

      if (gPref.prefHasUserValue("extensions.speeddial.thumbnail-" + originalDial + "-refreshinterval")) {
        gPref.setIntPref("extensions.speeddial.thumbnail-" + targetDial + "-refreshinterval", gPref.getIntPref("extensions.speeddial.thumbnail-" + originalDial + "-refreshinterval"));
      }

      if (gPref.prefHasUserValue("extensions.speeddial.thumbnail-" + originalDial + "-manualrefresh")) {
        gPref.setBoolPref("extensions.speeddial.thumbnail-" + targetDial + "-manualrefresh", gPref.getBoolPref("extensions.speeddial.thumbnail-" + originalDial + "-manualrefresh"));
      }
      
      if (moveIfPossible) {
        // Move image
        moveThumbnailImage(originalDial, targetDial);
        deassignThumbnail(originalDial);
      } else {
        // Copy image
        copyThumbnailImage(originalDial, targetDial);
      }
    } else {
      // We're inserting a new URL if possible...
      var url = transferUtils.retrieveURLFromData(targetData.data, targetData.contentType);
        
      if (!url || !url.length || url.indexOf(" ", 0) != -1 ||
        /^\s*(javascript|data):/.test(url))
        return;
      
      // Just add the information...
      gPref.setCharPref("extensions.speeddial.thumbnail-" + targetDial + "-url", url);

      var str = Components.classes["@mozilla.org/supports-string;1"].createInstance(Components.interfaces.nsISupportsString);
      str.data = "";
      gPref.setComplexValue("extensions.speeddial.thumbnail-" + targetDial + "-label", Components.interfaces.nsISupportsString, str);

      gPref.setCharPref("extensions.speeddial.thumbnail-" + targetDial + "-icon", "data:");
      gPref.setBoolPref("extensions.speeddial.thumbnail-" + targetDial + "-dynamictitle", true);

      // add time intervals
      var defaultRefreshInterval = gPref.getIntPref("extensions.speeddial.defaultRefreshInterval");
      if (defaultRefreshInterval > 0) {
        gPref.setIntPref("extensions.speeddial.thumbnail-" + targetDial + "-refreshinterval", defaultRefreshInterval);
      } else {
        if (gPref.prefHasUserValue("extensions.speeddial.thumbnail-" + targetDial + "-refreshinterval")) {
          gPref.clearUserPref("extensions.speeddial.thumbnail-" + targetDial + "-refreshinterval");
        }
      }

      if (gPref.prefHasUserValue("extensions.speeddial.thumbnail-" + targetDial + "-lastsaved")) {
        gPref.clearUserPref("extensions.speeddial.thumbnail-" + targetDial + "-lastsaved");
      }
    }
} catch (e) {
alert(e);
}
  },

  canHandleMultipleItems: false
};

function getSelectionFromXferData(aDragSession) {
    var newValue = {};
    var trans = Components.classes["@mozilla.org/widget/transferable;1"]
                          .createInstance(Components.interfaces.nsITransferable);
    trans.addDataFlavor("text/x-speeddial-entry");
    trans.addDataFlavor("text/x-moz-url");
    trans.addDataFlavor("text/x-moz-file");
    trans.addDataFlavor("text/unicode");
    var uri, extra, rSource, rParent, parent;
    if (aDragSession.numDropItems > 0) {
      var bestFlavour = {}, dataObj = {}, len = {};
      aDragSession.getData(trans, 0);
      trans.getAnyTransferData(bestFlavour, dataObj, len);
      dataObj = dataObj.value.QueryInterface(Components.interfaces.nsISupportsString);
      if (dataObj) {
        var value = dataObj.data.substring(0, len.value);

        newValue.value = value;
        switch (bestFlavour.value) {
        case "text/x-speeddial-entry":
          newValue.type = "speedDialId";
          break;
        case "text/x-moz-url":
        case "text/x-moz-file":
        case "text/unicode":
          newValue.type = "url";
          newValue.contentType = bestFlavour.value;
          newValue.data = dataObj.data;
          break;
        }
      }
    }
    return newValue;
}

function createTransferDataForEntry(targetEntry) {
  var value;
  var title;

  if (gPref.prefHasUserValue("extensions.speeddial.thumbnail-" + targetEntry + "-url")) {
    value = gPref.getCharPref("extensions.speeddial.thumbnail-" + targetEntry + "-url");
  } else {
    return null;
  }

  if (gPref.prefHasUserValue("extensions.speeddial.thumbnail-" + targetEntry + "-title")) {
    title = gPref.getCharPref("extensions.speeddial.thumbnail-" + targetEntry + "-title");
  } else {
    title = "";
  }

  if (title == "") {
    title = speedDialBundle.getString("untitled.label");
  }
  
  var urlString = value + "\n" + title;
  var htmlString = "<a href=\"" + value + "\">" + title + "</a>";

  var data = new TransferData();
  data.addDataForFlavour("text/x-speeddial-entry", targetEntry);
  if ((value == "") || (value == "about:blank")) {
    data.addDataForFlavour("text/unicode", value);
  } else {
    data.addDataForFlavour("text/x-moz-url", urlString);
    data.addDataForFlavour("text/unicode", value);
    data.addDataForFlavour("text/html", htmlString);
  }

  return data;
}
