var SpeedDial = {
  speedDialSlots: 9,
  prefs: null,
  bundle: null,
  THUMB_FOLDER: "SDThumbs",
  loadInNewWindow: false,
  loadInNewTab: false,
  showInTabContextMenu: false,
  showInAreaContextMenu: false,
  showInBookmarksMenu: false,
  overrideCtrlNumber: false,
  isSchedulerWindow: false,
  updateSchedules: null,
  prioritySchedules: null,
  scheduleTimer: null,
  processingPrioritySchedules: false,
  currentPrioritySchedule: -1,
  defaultRefreshInterval: 0,

  assignPopupShowing: function(event, showUnassigned) {
    var menu = event.target;
    // Empty current
    while (menu.firstChild)
      menu.removeChild(menu.firstChild);

    // Create new
    for (var c=1; c<=SpeedDial.speedDialSlots; c++) {
      if (showUnassigned || SpeedDial.prefs.prefHasUserValue("extensions.speeddial.thumbnail-" + c + "-url")) {
        var menuitem = document.createElement("menuitem");
        if (SpeedDial.prefs.prefHasUserValue("extensions.speeddial.thumbnail-" + c + "-url")) {
          var thumbnailTitle = SpeedDial.prefs.getComplexValue("extensions.speeddial.thumbnail-" + c + "-label", Components.interfaces.nsISupportsString).data;
          var thumbnailURL =  SpeedDial.prefs.getCharPref("extensions.speeddial.thumbnail-" + c + "-url");
          var thumbnailIcon =  SpeedDial.prefs.getCharPref("extensions.speeddial.thumbnail-" + c + "-icon");
          if (thumbnailTitle == "") {
            menuitem.setAttribute("label", "" + c + " - " + SpeedDial.bundle.getString("untitled.label"));
          } else {
            menuitem.setAttribute("label", "" + c + " - " + thumbnailTitle);
          }
          menuitem.setAttribute("tooltiptext", thumbnailURL);
          menuitem.setAttribute("image", thumbnailIcon);
        } else {
          menuitem.setAttribute("label", "" + c + " - " + SpeedDial.bundle.getString("unassigned.label"));
        }
        menuitem.setAttribute("accesskey", c);
        menuitem.setAttribute("speeddial", c);
        menuitem.setAttribute("class", "menuitem-iconic bookmark-item");
        menu.appendChild(menuitem);
      }
    }
    return true;
  },

  setAsSpeedDial: function(event, targetTab) {
    var speedDial = event.target.getAttribute("speeddial");
    if (!speedDial)
      return false;
    
    var browser;

    if (targetTab) {
      browser = targetTab.linkedBrowser;
    } else {
      browser = getBrowser().mCurrentBrowser;
    }

    if (SpeedDial.prefs.prefHasUserValue("extensions.speeddial.thumbnail-" + speedDial + "-url")) {
      if (!confirm(SpeedDial.bundle.getString("overwrite.warning"))) {
        return false;
      }
    }

    SpeedDial.prefs.setCharPref("extensions.speeddial.thumbnail-" + speedDial + "-icon", "");

    if (browser.mIconURL) {
      SpeedDial.loadIcon(browser.mIconURL, speedDial);
    }
    
    SpeedDial.prefs.setCharPref("extensions.speeddial.thumbnail-" + speedDial + "-url", browser.contentDocument.location);

    var str = Components.classes["@mozilla.org/supports-string;1"].createInstance(Components.interfaces.nsISupportsString);
    str.data = browser.contentDocument.title;
    SpeedDial.prefs.setComplexValue("extensions.speeddial.thumbnail-" + speedDial + "-label", Components.interfaces.nsISupportsString, str);

    SpeedDial.prefs.setBoolPref("extensions.speeddial.thumbnail-" + speedDial + "-dynamictitle", false);
    
    if (SpeedDial.defaultRefreshInterval > 0) {
      SpeedDial.prefs.setIntPref("extensions.speeddial.thumbnail-" + speedDial + "-refreshinterval", SpeedDial.defaultRefreshInterval);
    }

    // Take screenshot
    SpeedDial.saveSnapshot(browser, speedDial);

    return true;
  },

  showWelcomeScreenIfNeeded: function() {
     var currentVersion = SpeedDial.prefs.getCharPref("extensions.speeddial.currentVersion");

     if (currentVersion != "0.3.4") {
       SpeedDial.prefs.setCharPref("extensions.speeddial.currentVersion", "0.3.4");
       var targetURL;
       if (currentVersion == "none") {
         targetURL = "http://speeddial.uworks.net/welcome.html";
       } else {
         targetURL = "http://speeddial.uworks.net/updated.html";
       }
       var newTab = getBrowser().addTab(targetURL);
       getBrowser().selectedTab = newTab;
     }
  },


  loadIcon: function(url, entry) {
    var IOSVC = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService);
    var chan = IOSVC.newChannel(url, null, null);
    var listener = new SpeedDial.SpeedDialIconLoadListener(chan, entry);
    chan.notificationCallbacks = listener;
    chan.asyncOpen(listener, null);
  },

  openSpeedDialCheckMiddleClick: function(node, event) {
    if (event.button == 1) {
      var func = new Function("event",node.getAttribute("oncommand"));
      func.call(node,event);
      if ("tagName" in node) {
        if (node.namespaceURI == "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul" && (node.tagName == "menupopup" || node.tagName == "popup"))
          node.hidePopup();
      }
    }
  },

  openSpeedDial: function(event) {
    var targetSpeedDial = event.target.getAttribute("speeddial");

    if (!SpeedDial.prefs.prefHasUserValue("extensions.speeddial.thumbnail-" + targetSpeedDial + "-url"))
      return;

    var thumbnailURL =  SpeedDial.prefs.getCharPref("extensions.speeddial.thumbnail-" + targetSpeedDial + "-url");
    var where = whereToOpenLink(event);
    openUILinkIn(thumbnailURL, where);
  },

  setSpeedDialIcon: function(data, speedDial) {
    SpeedDial.prefs.setCharPref("extensions.speeddial.thumbnail-" + speedDial + "-icon", data);
  },

  saveSnapshot: function(browser, speedDial) {
    // Verify folder
    SpeedDial.verifySpeedDialThumbFolder();

    var currentCanvas = document.createElementNS("http://www.w3.org/1999/xhtml", "canvas");
    currentCanvas.width = SpeedDial.prefs.getIntPref("extensions.speeddial.thumbnailImageWidth");
    currentCanvas.height = SpeedDial.prefs.getIntPref("extensions.speeddial.thumbnailImageHeight");

    var ctx = currentCanvas.getContext("2d");
    
    ctx.save();
    ctx.scale(currentCanvas.width/browser.contentWindow.innerWidth, currentCanvas.width/browser.contentWindow.innerWidth);

    // bug 313178: when painting thumbnails without scroll, there's an important bug
    if (browser.contentWindow.scrollY != 0) {
      ctx.drawWindow(browser.contentWindow, 
          0, 
          0, 
          browser.contentWindow.innerWidth, 
          browser.contentWindow.innerHeight + browser.contentWindow.scrollMaxY, 
          "rgb(255,255,255)");
    } else {
      ctx.drawWindow(browser.contentWindow, 
          0, 
          0, 
          browser.contentWindow.innerWidth, 
          browser.contentWindow.innerHeight, 
          "rgb(255,255,255)");
    }
    ctx.restore();

    var saveTime = (new Date()).getTime();

    var file = Components.classes["@mozilla.org/file/directory_service;1"]
                     .getService(Components.interfaces.nsIProperties)
                     .get("ProfD", Components.interfaces.nsIFile);
    file.append(SpeedDial.THUMB_FOLDER);
    file.append("thumbnail-" + speedDial + "-" + saveTime + ".png");

    SpeedDial.saveCanvas(currentCanvas, file);

    // If there's a previous one, delete it
    
    SpeedDial.removeThumbnailImage(speedDial);

    // Save preference, we've updated the thumbnail
    SpeedDial.prefs.setCharPref("extensions.speeddial.thumbnail-" + speedDial + "-lastsaved", saveTime);
  },

  removeThumbnailImage: function(thumbNumber) {
    var file = Components.classes["@mozilla.org/file/directory_service;1"]
                       .getService(Components.interfaces.nsIProperties)
                       .get("ProfD", Components.interfaces.nsIFile);
    file.append(SpeedDial.THUMB_FOLDER);
    if (SpeedDial.prefs.prefHasUserValue("extensions.speeddial.thumbnail-" + thumbNumber + "-lastsaved")) {
      file.append("thumbnail-" + thumbNumber + "-" + SpeedDial.prefs.getCharPref("extensions.speeddial.thumbnail-" + thumbNumber + "-lastsaved") + ".png");
    } else {
      file.append("thumbnail-" + thumbNumber + ".png");
    }
    if (file.exists() && !file.isDirectory()) {
      try {
        file.remove(false);
      } catch(e) {
        // Do nothing...
        // TODO: retry later?
      }
    }
  },

  addSpeedDialTabContextMenu: function() {
    // get strings 
    var menuLabel = SpeedDial.bundle.getString("speeddialCmd.label");
    var menuAccessKey = SpeedDial.bundle.getString("speeddialCmd.accesskey");

    // create new menu item
    var speedDialTabItem = document.createElement("menu");
    speedDialTabItem.setAttribute("id", "tabContextSpeedDial");
    speedDialTabItem.setAttribute("label", menuLabel);
    speedDialTabItem.setAttribute("accesskey", menuAccessKey);
    var speedDialTabMenu = document.createElement("menupopup");
    speedDialTabMenu.setAttribute("onpopupshowing", "return SpeedDial.assignPopupShowing(event, true);");
    speedDialTabMenu.setAttribute("oncommand", "SpeedDial.setAsSpeedDial(event,document.popupNode);");
    speedDialTabItem.appendChild(speedDialTabMenu);

    // add to tab context menu
    var tabbrowser = getBrowser();
    var tabMenu = document.getAnonymousElementByAttribute(tabbrowser,"anonid","tabContextMenu");

    var insertPos = document.getElementById("tabContextUndoCloseTab");

    if (!insertPos) {
      insertPos = tabMenu.lastChild.previousSibling;
    }
    tabMenu.insertBefore(speedDialTabItem, insertPos);

    speedDialTabItem.hidden = !SpeedDial.showInTabContextMenu;
  },

  verifySpeedDialThumbFolder: function() {
    var file = Components.classes["@mozilla.org/file/directory_service;1"]
                     .getService(Components.interfaces.nsIProperties)
                     .get("ProfD", Components.interfaces.nsIFile);
    file.append(SpeedDial.THUMB_FOLDER);
    if( !file.exists() || !file.isDirectory() ) {   // if it doesn't exist, create
      file.create(Components.interfaces.nsIFile.DIRECTORY_TYPE, 0755);
    }
  },

  saveCanvas: function(canvas, destFile) {
try {
    // convert string filepath to an nsIFile
    var file = Components.classes["@mozilla.org/file/local;1"]
                       .createInstance(Components.interfaces.nsILocalFile);

    // create a data url from the canvas and then create URIs of the source and targets  
    var io = Components.classes["@mozilla.org/network/io-service;1"]
                       .getService(Components.interfaces.nsIIOService);
    var source = io.newURI(canvas.toDataURL("image/png", ""), "UTF8", null);
    var target = io.newFileURI(destFile)
  
    // prepare to save the canvas data
    var persist = Components.classes["@mozilla.org/embedding/browser/nsWebBrowserPersist;1"]
                          .createInstance(Components.interfaces.nsIWebBrowserPersist);
  
    persist.persistFlags = Components.interfaces.nsIWebBrowserPersist.PERSIST_FLAGS_REPLACE_EXISTING_FILES;
    persist.persistFlags |= Components.interfaces.nsIWebBrowserPersist.PERSIST_FLAGS_AUTODETECT_APPLY_CONVERSION;
  
    // displays a download dialog (remove these 3 lines for silent download)
/*
    var xfer = Components.classes["@mozilla.org/transfer;1"]
                       .createInstance(Components.interfaces.nsITransfer);
    xfer.init(source, target, "", null, null, null, persist);
    persist.progressListener = xfer;
*/
  
    // save the canvas data to the file
    persist.saveURI(source, null, null, null, null, destFile);
} catch (e) {
  alert(e);
}
  },

  newTabListener: function(event) {
    if (SpeedDial.loadInNewTab)
      setTimeout(SpeedDial.checkNewTab, 100, event.originalTarget, 10);
  },

  contentAreaAttrListener: function(event) {
    if (!SpeedDial.showInAreaContextMenu) {
      document.getElementById("speeddialContext").hidden = true;
      return;
    }

    if (event.attrName == "hidden") {
      if (event.attrChange != event.REMOVAL) {
        document.getElementById("speeddialContext").hidden = event.newValue;
      } else {
        document.getElementById("speeddialContext").hidden = false;
      }
    }
  },

  checkNewTab: function(tab, attempts) {
    if (!tab.hasAttribute("busy")) {
      var domWindow = tab.linkedBrowser.docShell
                            .QueryInterface(Components.interfaces.nsIInterfaceRequestor)
                            .getInterface(Components.interfaces.nsIDOMWindow);
      var webNav = domWindow.QueryInterface(Components.interfaces.nsIInterfaceRequestor)
                            .getInterface(Components.interfaces.nsIWebNavigation);

      if ((!webNav || !webNav.currentURI) && attempts) {
        window.setTimeout(SpeedDial.checkNewTab, 100, tab, --attempts);
        return;
      } 
      if ((webNav.currentURI.spec == "about:blank") && !webNav.canGoBack && !webNav.canGoForward)
        webNav.loadURI("chrome://speeddial/content", null, null, null, null);
    }
  },

  checkNewBrowser: function(attempts) {
    var domWindow = getBrowser().mCurrentBrowser.docShell
                            .QueryInterface(Components.interfaces.nsIInterfaceRequestor)
                            .getInterface(Components.interfaces.nsIDOMWindow);
    var webNav = domWindow.QueryInterface(Components.interfaces.nsIInterfaceRequestor)
                            .getInterface(Components.interfaces.nsIWebNavigation);

    if (!webNav.currentURI && attempts) {
      window.setTimeout(SpeedDial.checkNewBrowser, 100, --attempts);
      return;
    }

    if (gIsLoadingBlank && (getBrowser().mTabs.length < 2) && (webNav.currentURI.spec == "about:blank") && !webNav.canGoBack && !webNav.canGoForward && !getBrowser().mCurrentTab.hasAttribute("busy"))
      webNav.loadURI("chrome://speeddial/content", null, null, null, null);
  },

  ctrlNumberSpeedDial: function (event) {
    if (event.altKey && event.keyCode == KeyEvent.DOM_VK_RETURN) {
      // XXXblake Proper fix is to just check whether focus is in the urlbar. However, focus with the autocomplete widget is all
      // hacky and broken and there's no way to do that right now. So this just patches it to ensure that alt+enter works when focus
      // is on a link.
      if (!(document.commandDispatcher.focusedElement instanceof HTMLAnchorElement)) {
        // Don't let winxp beep on ALT+ENTER, since the URL bar uses it.
        event.preventDefault();
        return;
      }
    }

    if (!event.ctrlKey)
      return;

    var index = event.charCode - 48;
    if (index < 1 || index > 9)
    return;

    if (!SpeedDial.prefs.prefHasUserValue("extensions.speeddial.thumbnail-" + index + "-url"))
      return;

    var thumbnailURL =  SpeedDial.prefs.getCharPref("extensions.speeddial.thumbnail-" + index + "-url");
    if (thumbnailURL) {
      openUILinkIn(thumbnailURL, "current");
    } else {
      return;
    }

    event.preventDefault();
    event.preventBubble();
    event.preventCapture();
    event.stopPropagation();
  },

  init: function () {
    try {
    SpeedDial.prefs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefBranch2);
    SpeedDial.bundle = document.getElementById("bundle_speeddial");

    // Load settings
    SpeedDial.loadInNewWindow = SpeedDial.prefs.getBoolPref("extensions.speeddial.loadInNewWindow");
    SpeedDial.loadInNewTab = SpeedDial.prefs.getBoolPref("extensions.speeddial.loadInNewTab");
    SpeedDial.showInTabContextMenu = SpeedDial.prefs.getBoolPref("extensions.speeddial.showInTabContextMenu");
    SpeedDial.showInAreaContextMenu = SpeedDial.prefs.getBoolPref("extensions.speeddial.showInAreaContextMenu");
    SpeedDial.showInBookmarksMenu = SpeedDial.prefs.getBoolPref("extensions.speeddial.showInBookmarksMenu");
    SpeedDial.overrideCtrlNumber = SpeedDial.prefs.getBoolPref("extensions.speeddial.overrideCtrlNumber");
    SpeedDial.defaultRefreshInterval = SpeedDial.prefs.getIntPref("extensions.speeddial.defaultRefreshInterval");

    // Register listeners
    var bookmarkContextItem = document.getElementById("context-bookmarkpage");
    bookmarkContextItem.addEventListener("DOMAttrModified", SpeedDial.contentAreaAttrListener, false);
    document.addEventListener("TabOpen", SpeedDial.newTabListener, false);
    SpeedDialPrefObserver.addPrefObserver();

    // Register window observer
    var watcherService =
      Components.classes["@mozilla.org/embedcomp/window-watcher;1"]
        .getService(Components.interfaces.nsIWindowWatcher);
    watcherService.registerNotification(SpeedDialWindowObserver);

    // Check if this window is the scheduler
    SpeedDial.checkWindow();

    // Hide some things
    if (!SpeedDial.showInAreaContextMenu ) {
      document.getElementById("speeddialContext").hidden = true;
    }
    if (!SpeedDial.showInBookmarksMenu) {
      document.getElementById("speeddialBookmarksMenu").hidden = true;
    }

    // Add to tab context menu
    setTimeout(SpeedDial.addSpeedDialTabContextMenu, 100);

    // Substitute ctrl+number
    if (SpeedDial.overrideCtrlNumber) {
      window.originalCtrlNumberTabSelection = window.ctrlNumberTabSelection;
      window.ctrlNumberTabSelection = SpeedDial.ctrlNumberSpeedDial;
    }

    // Check current window...
    if (SpeedDial.loadInNewWindow)
      setTimeout(SpeedDial.checkNewBrowser, 100, 10);
    } catch (e) {
      alert(e);
    }

    setTimeout(function() {SpeedDial.showWelcomeScreenIfNeeded() }, 400);
  },

  unload: function () {
    // Unregister listeners
    var bookmarkContextItem = document.getElementById("context-bookmarkpage");
    bookmarkContextItem.removeEventListener("DOMAttrModified", SpeedDial.contentAreaAttrListener, false);
    document.removeEventListener("TabOpen", SpeedDial.newTabListener, false);
    SpeedDialPrefObserver.removePrefObserver();

    // Unregister window observer
    watcherService.unregisterNotification(SpeedDialWindowObserver);

    if (SpeedDial.isSchedulerWindow) {
      SpeedDial.stopScheduling();
    }
  },

  registerBackgroundLoader: function() {
    var loaderBrowser = document.getElementById("speedDialLoaderBrowser");
    loaderBrowser.addProgressListener(SpeedDialBackgroundBrowserListener, Components.interfaces.nsIWebProgress.NOTIFY_STATE_WINDOW);
    loaderBrowser.addEventListener("DOMLinkAdded", SpeedDialBackgroundBrowserListener, true);
  },

  unregisterBackgroundLoader: function() {
    var loaderBrowser = document.getElementById("speedDialLoaderBrowser");
    loaderBrowser.removeProgressListener(SpeedDialBackgroundBrowserListener);
    loaderBrowser.removeEventListener("DOMLinkAdded", SpeedDialBackgroundBrowserListener, true);
  },

  startBackgroundLoad: function() {
    var loaderBrowser = document.getElementById("speedDialLoaderBrowser");
    var backgroundBrowserWidth = SpeedDial.prefs.getIntPref("extensions.speeddial.backgroundBrowserWidth");
    var backgroundBrowserHeight = SpeedDial.prefs.getIntPref("extensions.speeddial.backgroundBrowserHeight");

    loaderBrowser.setAttribute("width", backgroundBrowserWidth);
    loaderBrowser.setAttribute("minwidth", backgroundBrowserWidth);
    loaderBrowser.setAttribute("maxwidth", backgroundBrowserWidth);
    loaderBrowser.setAttribute("height", backgroundBrowserHeight);
    loaderBrowser.setAttribute("minheight", backgroundBrowserHeight);
    loaderBrowser.setAttribute("maxheight", backgroundBrowserHeight);

    loaderBrowser.style.width = backgroundBrowserWidth + "px !important";
    loaderBrowser.style.maxWidth = loaderBrowser.style.width;
    loaderBrowser.style.minWidth = loaderBrowser.style.width;
    loaderBrowser.style.height = backgroundBrowserHeight + "px !important";
    loaderBrowser.style.maxHeight = loaderBrowser.style.height;
    loaderBrowser.style.minHeight = loaderBrowser.style.height;

    loaderBrowser.mIconURL = null;
  
    loaderBrowser.setAttribute("src", SpeedDial.prefs.getCharPref("extensions.speeddial.thumbnail-" + SpeedDial.currentPrioritySchedule + "-url"));
  },

  backgroundIconLoaded: function(url) {
    var loaderBrowser = document.getElementById("speedDialLoaderBrowser");
    loaderBrowser.mIconURL = url;
    SpeedDial.loadIcon(url, SpeedDial.currentPrioritySchedule);
  },

  backgroundLoadFinished: function(event) {
    var loaderBrowser = document.getElementById("speedDialLoaderBrowser");
    if (SpeedDial.currentPrioritySchedule > -1) {
      try {
        // Get the IOService so we can make URIs
        const ioService = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService);
        var origURI = ioService.newURI(loaderBrowser.contentDocument.documentURI, loaderBrowser.contentDocument.characterSet, null);

        if (origURI.spec.substr(0, "about:neterror".length) == "about:neterror") {
alert("error in load");
          if (SpeedDial.prefs.prefHasUserValue("extensions.speeddial.thumbnail-" + SpeedDial.currentPrioritySchedule + "-lastsave")) {
          } else {
            // It's the first time we've loaded the page.. place some error information
            if (SpeedDial.prefs.prefHasUserValue("extensions.speeddial.thumbnail-" + SpeedDial.currentPrioritySchedule + "-dynamictitle") &&
                SpeedDial.prefs.getBoolPref("extensions.speeddial.thumbnail-" + SpeedDial.currentPrioritySchedule + "-dynamictitle")) {
                
            }
          }
        } else {
          // Check if we've to get default icon
          if (!loaderBrowser.mIconURL) {
            if (("schemeIs" in loaderBrowser.currentURI) && (loaderBrowser.currentURI.schemeIs("http") || loaderBrowser.currentURI.schemeIs("https"))) {
              var url = loaderBrowser.currentURI.prePath + "/favicon.ico";
              SpeedDial.loadIcon(url, SpeedDial.currentPrioritySchedule);
            }
          }

          // Check if dynamic title...
          if (SpeedDial.prefs.prefHasUserValue("extensions.speeddial.thumbnail-" + SpeedDial.currentPrioritySchedule + "-dynamictitle") &&
               SpeedDial.prefs.getBoolPref("extensions.speeddial.thumbnail-" + SpeedDial.currentPrioritySchedule + "-dynamictitle")) {
            var str = Components.classes["@mozilla.org/supports-string;1"].createInstance(Components.interfaces.nsISupportsString);
            str.data = loaderBrowser.contentDocument.title;
            SpeedDial.prefs.setComplexValue("extensions.speeddial.thumbnail-" + SpeedDial.currentPrioritySchedule + "-label", Components.interfaces.nsISupportsString, str);
          }

          SpeedDial.saveSnapshot(loaderBrowser, SpeedDial.currentPrioritySchedule);
        }

        // Remove manual refresh preference if necessary
        if (SpeedDial.prefs.prefHasUserValue("extensions.speeddial.thumbnail-" + SpeedDial.currentPrioritySchedule + "-manualrefresh")) {
          SpeedDial.prefs.clearUserPref("extensions.speeddial.thumbnail-" + SpeedDial.currentPrioritySchedule + "-manualrefresh");
        }

        SpeedDial.prioritySchedules.shift();
        SpeedDial.currentPrioritySchedule = -1;
        loaderBrowser.setAttribute("src", "about:blank");
      } catch (e) {
        alert("backgroundLoadFinished: " + e);
      }
    } else if (SpeedDial.processingPrioritySchedules) {
      if (loaderBrowser.contentDocument.location == "about:blank") {
        SpeedDial.processingPrioritySchedules = false;
        SpeedDial.processPrioritySchedules();
      } else {
        loaderBrowser.setAttribute("src", "about:blank");
      }
    }
  },

  checkWindow: function() {
    if (SpeedDial.isSchedulerWindow) return;

    var wm = Components.classes['@mozilla.org/appshell/window-mediator;1'].getService(Components.interfaces.nsIWindowMediator);
    var windowIter = wm.getEnumerator('navigator:browser');
   
    if (windowIter.hasMoreElements() && (window == windowIter.getNext())) {
      // We're the scheduler window now!
      SpeedDial.startScheduling();
    }
  },

  startScheduling: function() {
    SpeedDial.isSchedulerWindow = true;

    SpeedDial.registerBackgroundLoader();

    SpeedDial.scheduleTimer = Components.classes["@mozilla.org/timer;1"].createInstance(Components.interfaces.nsITimer);
    SpeedDial.prioritySchedules = new Array();
    SpeedDial.updateSchedules = new Array();
    
    setTimeout(SpeedDial.checkThumbnailRefreshing, 100);
  },

  stopScheduling: function() {
    SpeedDial.isSchedulerWindow = false;

    SpeedDial.unregisterBackgroundLoader();

    SpeedDial.scheduleTimer.cancel();
    SpeedDial.scheduleTimer = null;
  },

  checkThumbnailRefreshing: function() {
//alert("checkThumbnailRefreshing");
    // Cancel current updates
    SpeedDial.scheduleTimer.cancel();

    SpeedDial.updateSchedules.length = 0;

    // Determine update and priority schedules
    for (var c=1; c<=SpeedDial.speedDialSlots; c++) {
      if (SpeedDial.prefs.prefHasUserValue("extensions.speeddial.thumbnail-" + c + "-url")) {
        if (!SpeedDial.prefs.prefHasUserValue("extensions.speeddial.thumbnail-" + c + "-lastsaved") || SpeedDial.prefs.prefHasUserValue("extensions.speeddial.thumbnail-" + c + "-manualrefresh")) {
          SpeedDial.addPrioritySchedule(c);
        } else if (SpeedDial.prefs.prefHasUserValue("extensions.speeddial.thumbnail-" + c + "-refreshinterval")) {
          SpeedDial.updateSchedules.push(c);
        }
      }
    }
//alert("processed schedules: priority " + SpeedDial.prioritySchedules.length + ", updates " + SpeedDial.updateSchedules.length);

    if (SpeedDial.updateSchedules.length > 0) {
      // Calculate wait time
      var smallestWait = 0;
      var currentTime = (new Date()).getTime();
      
      for (var c=0; c<SpeedDial.updateSchedules.length; c++) {
        var lastSaved = parseInt(SpeedDial.prefs.getCharPref("extensions.speeddial.thumbnail-" + SpeedDial.updateSchedules[c] + "-lastsaved"));
        var currentWait = lastSaved + SpeedDial.prefs.getIntPref("extensions.speeddial.thumbnail-" + SpeedDial.updateSchedules[c] + "-refreshinterval") * 1000 - currentTime;
        
        if (currentWait < smallestWait) smallestWait = currentWait;
      }

      if (smallestWait < 100) smallestWait = 100;

      SpeedDial.scheduleTimer.initWithCallback(SpeedDialTimerListener, smallestWait, Components.interfaces.nsITimer.TYPE_ONE_SHOT);
    }
  },

  addPrioritySchedule: function(targetDial) {
    if (SpeedDial.prioritySchedules.indexOf(targetDial) < 0) {
      SpeedDial.prioritySchedules.push(targetDial);
      SpeedDial.processPrioritySchedules();
    }
  },

  processPrioritySchedules: function() {
    if (SpeedDial.prioritySchedules.length > 0) {
      SpeedDial.currentPrioritySchedule = SpeedDial.prioritySchedules[0];
      if (!SpeedDial.processingPrioritySchedules) {
        SpeedDial.processingPrioritySchedules = true;
        SpeedDial.startBackgroundLoad();
      }
    } else {
      SpeedDial.processingPrioritySchedules = false;
    }
  },

  checkSchedule: function() {
     if (SpeedDial.updateSchedules.length < 1) return;

     // Find our target
     var smallestWait = 0;
     var currentTime = (new Date()).getTime();
     
     for (var c=0; c<SpeedDial.updateSchedules.length; c++) {
       var targetIndex = SpeedDial.updateSchedules[c];
       if (SpeedDial.prioritySchedules.indexOf(targetIndex) < 0) {
         var lastSaved = parseInt(SpeedDial.prefs.getCharPref("extensions.speeddial.thumbnail-" + targetIndex + "-lastsaved"));
         var currentWait = lastSaved + SpeedDial.prefs.getIntPref("extensions.speeddial.thumbnail-" + targetIndex + "-refreshinterval") * 1000 - currentTime;
         if (currentWait < 0) {
           SpeedDial.addPrioritySchedule(targetIndex);
         } else if (currentWait < smallestWait) {
           smallestWait = currentWait;
         }
       }
     }

     if (smallestWait > 100) {
       SpeedDial.scheduleTimer.initWithCallback(SpeedDialTimerListener, smallestWait, Components.interfaces.nsITimer.TYPE_ONE_SHOT);
     } else {
       // Re-check again
       SpeedDial.scheduleTimer.initWithCallback(SpeedDialTimerListener, 100, Components.interfaces.nsITimer.TYPE_ONE_SHOT);
     }
  },

  isBackgroundLoaderBusy: function() {
    // TODO: implement multiple browsers to do parallel background loading... future feature
    // It should show it's busy when all parallel browsers are working
    return SpeedDial.processingPrioritySchedules;
  },

  SpeedDialIconLoadListener: function(channel, speedDial) {
    this.mChannel = channel;
    this.mData = "";
    this.mSpeedDial = speedDial;
  }
};

SpeedDial.SpeedDialIconLoadListener.prototype =
{
    mCountRead: null,
    mChannel: null,
    mBytes: Array (),
    mStream: null,
    mData: null,
    mSpeedDial: -1,
    
    QueryInterface: function (iid)
    {
        if (!iid.equals(Components.interfaces.nsISupports) &&
            !iid.equals(Components.interfaces.nsIInterfaceRequestor) &&
            !iid.equals(Components.interfaces.nsIRequestObserver) &&
            !iid.equals(Components.interfaces.nsIStreamListener)) {
            throw Components.results.NS_ERROR_NO_INTERFACE;
        }
        return this;
    },
    
    getInterface: function(iid)
    {
        return this.QueryInterface(iid);
    },
    
    onStartRequest: function (aRequest, aContext)
    {
        this.mStream = Components.classes['@mozilla.org/binaryinputstream;1'].createInstance(Components.interfaces.nsIBinaryInputStream);
    },
    
    onStopRequest: function (aRequest, aContext, aStatusCode)
    {
       this.mData = "data:" + this.mChannel.contentType + ";base64," + btoa(String.fromCharCode.apply(null, this.mBytes));
	 SpeedDial.setSpeedDialIcon(this.mData, this.mSpeedDial);
       this. mChannel = null;
    },
    
    onDataAvailable: function (aRequest, aContext, aInputStream, aOffset, aCount)
    {
        this. mStream. setInputStream (aInputStream);
        var chunk = this. mStream. readByteArray (aCount);
        this. mBytes = this. mBytes. concat (chunk);
    }
}

var SpeedDialPrefObserver = 
{
  updateTimer: null,

  prefObserver : {
    observe: function(subject, topic, data) {
      // subject is the nsIPrefBranch we're observing (after appropriate QI)
      // data is the name of the pref that's been changed (relative to subject)
      if (topic == "nsPref:changed") {
        if (data.substr(0, "extensions.speeddial.thumbnail-".length) == "extensions.speeddial.thumbnail-") {
          // A thumbnail has been updated!
          if (SpeedDialPrefObserver.updateTimer == null) {
            SpeedDialPrefObserver.updateTimer = Components.classes["@mozilla.org/timer;1"].createInstance(Components.interfaces.nsITimer);
          }
          SpeedDialPrefObserver.updateTimer.cancel();
          SpeedDialPrefObserver.updateTimer.initWithCallback(SpeedDialPrefObserver.prefObserver, 100, Components.interfaces.nsITimer.TYPE_ONE_SHOT);
        } else {
          switch (data) {
            case "extensions.speeddial.loadInNewWindow":
              SpeedDial.loadInNewWindow = SpeedDial.prefs.getBoolPref("extensions.speeddial.loadInNewWindow");
              break;
            case "extensions.speeddial.loadInNewTab":
              SpeedDial.loadInNewTab = SpeedDial.prefs.getBoolPref("extensions.speeddial.loadInNewTab");
              break;
            case "extensions.speeddial.showInTabContextMenu":
              SpeedDial.showInTabContextMenu = SpeedDial.prefs.getBoolPref("extensions.speeddial.showInTabContextMenu");
              document.getElementById("tabContextSpeedDial").hidden = !SpeedDial.showInTabContextMenu;
              break;
            case "extensions.speeddial.showInAreaContextMenu":
              SpeedDial.showInAreaContextMenu = SpeedDial.prefs.getBoolPref("extensions.speeddial.showInAreaContextMenu");
              document.getElementById("speeddialContext").hidden = !SpeedDial.showInAreaContextMenu;
              break;
            case "extensions.speeddial.showInBookmarksMenu":
              SpeedDial.showInBookmarksMenu = SpeedDial.prefs.getBoolPref("extensions.speeddial.showInBookmarksMenu");
              document.getElementById("speeddialBookmarksMenu").hidden = !SpeedDial.showInBookmarksMenu;
              break;
            case "extensions.speeddial.overrideCtrlNumber":
              SpeedDial.overrideCtrlNumber = SpeedDial.prefs.getBoolPref("extensions.speeddial.overrideCtrlNumber");

              window.removeEventListener("keypress", window.ctrlNumberTabSelection, false);

              if (SpeedDial.overrideCtrlNumber) {
                if (!window.originalCtrlNumberTabSelection)
                  window.originalCtrlNumberTabSelection = window.ctrlNumberTabSelection;
                window.ctrlNumberTabSelection = SpeedDial.ctrlNumberSpeedDial;
              } else {
                if (window.originalCtrlNumberTabSelection)
                  window.ctrlNumberTabSelection = window.originalCtrlNumberTabSelection;
              }

              window.addEventListener("keypress", window.ctrlNumberTabSelection, false);
              break;
            case "extensions.speeddial.defaultRefreshInterval":
              SpeedDial.defaultRefreshInterval = SpeedDial.prefs.getIntPref("extensions.speeddial.defaultRefreshInterval");
              break;
          }
        }
      }
    },

    QueryInterface : function (aIID) {
      if (aIID.equals(Components.interfaces.nsIObserver) || 
      aIID.equals(Components.interfaces.nsITimerCallback) ||
      aIID.equals(Components.interfaces.nsISupports) ||
      aIID.equals(Components.interfaces.nsISupportsWeakReference))
        return this;
      throw Components.results.NS_NOINTERFACE;
    },

    notify: function(timer) {
      SpeedDial.checkThumbnailRefreshing();
    }
  },

  addPrefObserver : function () {
    if (!SpeedDial.prefs) return;
    SpeedDial.prefs.addObserver("extensions.speeddial.", SpeedDialPrefObserver.prefObserver, true);
  },

  removePrefObserver : function () {
    if (!SpeedDial.prefs) return;
    SpeedDial.prefs.removeObserver("extensions.speeddial.", SpeedDialPrefObserver.prefObserver);
  }
};

// Window close observer
var SpeedDialWindowObserver = {
  observe: function(subject,topic,data){
    switch(topic){
      case 'domwindowclosed':
        setTimeout(SpeedDial.checkWindow, 0);
        break;
    }
  }
};

// Thumbnail scheduler
var SpeedDialTimerListener = {
    QueryInterface : function (aIID) {
      if (aIID.equals(Components.interfaces.nsITimerCallback) ||
      aIID.equals(Components.interfaces.nsISupports) ||
      aIID.equals(Components.interfaces.nsISupportsWeakReference))
        return this;
      throw Components.results.NS_NOINTERFACE;
    },

    notify: function(timer) {
      if (!SpeedDial.isSchedulerWindow) return;

      SpeedDial.checkSchedule();
    }
};

var SpeedDialBackgroundBrowserListener =
{
  inhibitNextStop: false,

  QueryInterface: function(aIID)
  {
   if (aIID.equals(Components.interfaces.nsIWebProgressListener) ||
       aIID.equals(Components.interfaces.nsIDOMEventListener) ||
       aIID.equals(Components.interfaces.nsISupportsWeakReference) ||
       aIID.equals(Components.interfaces.nsISupports))
     return this;
   throw Components.results.NS_NOINTERFACE;
  },

  onStateChange: function(aProgress, aRequest, aFlag, aStatus)
  {
//   if(aFlag & Components.interfaces.nsIWebProgressListener.STATE_START)
   var loaderBrowser = document.getElementById("speedDialLoaderBrowser");
   if (aProgress.DOMWindow == loaderBrowser.contentWindow) {
     //if (aFlag & Components.interfaces.nsIWebProgressListener.STATE_REDIRECTING) {
     //  SpeedDialBackgroundBrowserListener.inhibitNextStop = true;
     //} else 
     if ((aFlag & Components.interfaces.nsIWebProgressListener.STATE_STOP) && (aFlag & Components.interfaces.nsIWebProgressListener.STATE_IS_NETWORK)) {
       if (SpeedDialBackgroundBrowserListener.inhibitNextStop) {
         SpeedDialBackgroundBrowserListener.inhibitNextStop = false;
       } else {
         // This fires when the load finishes
         setTimeout(SpeedDial.backgroundLoadFinished, 100);
       }
     }
   }
   return 0;
  },

  onLocationChange: function(aProgress, aRequest, aURI)
  {
   // This fires when the location bar changes i.e load event is confirmed
   // or when the user switches tabs
   return 0;
  },

  // For definitions of the remaining functions see XulPlanet.com
  onProgressChange: function() {return 0;},
  onStatusChange: function() {return 0;},
  onSecurityChange: function() {return 0;},
  onLinkIconAvailable: function() {return 0;},
  
  handleEvent: function(event) {
    if (!event.originalTarget.rel.match((/(?:^|\s)icon(?:\s|$)/i)))
      return;

    // We have an icon.
    var href = event.originalTarget.href;
    if (!href)
      return;

    SpeedDial.backgroundIconLoaded(href);
  }
}


window.addEventListener("load", function(e) { SpeedDial.init(); }, false);
window.addEventListener("unload", function(e) { SpeedDial.unload(); }, false);
